@@ -27,6 +27,8 @@ module Map : sig
   module type S = sig
     include Map.S
     include Show1 with type 'a t := 'a t
+    val pp_key : Format.formatter -> key -> unit
+    val show_key : key -> string
   end
 
   module type OrderedType = sig
@@ -88,6 +90,7 @@ end
 
 module Loc : sig
   type t = {
+    client_payload : Obj.t option;
     source_name : string;
     source_start: int;
     source_stop: int;
@@ -99,7 +102,12 @@ module Loc : sig
   val equal : t -> t -> bool
   val compare : t -> t -> int
 
-  val initial : string -> t
+  val initial : ?client_payload:Obj.t -> string -> t
+  (* merge left right *)
+  val merge : ?merge_payload:(Obj.t option -> Obj.t option -> Obj.t option) -> t -> t -> t
+  (* starts/end n chars before/after*)
+  val extend : int -> t -> t
+
 end
 
 (******************** list ******************)
@@ -113,7 +121,14 @@ val for_all2 : ('a -> 'a -> bool) -> 'a list -> 'a list -> bool
 val for_all23 :  argsdepth:int -> (argsdepth:int -> matching:bool -> 'x -> 'y -> 'z -> 'a -> 'a -> bool) -> 'x -> 'y -> 'z -> 'a list -> 'a list -> bool
 val for_all3b : ('a -> 'a -> bool -> bool) -> 'a list -> 'a list -> bool list -> bool -> bool
 type arg_mode = Input | Output
-val for_all3b3 : argsdepth:int -> (argsdepth:int -> matching:bool -> 'x -> 'y -> 'z -> 'a -> 'a -> bool) -> 'x -> 'y -> 'z -> 'a list -> 'a list -> arg_mode list -> bool -> bool
+[@@deriving show, ord]
+
+type mode_aux =
+  | Fo of arg_mode
+  | Ho of arg_mode * mode
+and mode = mode_aux list
+
+val for_all3b3 : argsdepth:int -> (argsdepth:int -> matching:bool -> 'x -> 'y -> 'z -> 'a -> 'a -> bool) -> 'x -> 'y -> 'z -> 'a list -> 'a list -> mode -> bool -> bool
 (*uses physical equality and calls anomaly if the element is not in the list*)
 val remove_from_list : 'a -> 'a list -> 'a list
 (* returns Some t where f x = Some t for the first x in the list s.t.
@@ -132,6 +147,7 @@ val uniq : 'a list -> 'a list
 
 val option_get : ?err:string -> 'a option -> 'a
 val option_map : ('a -> 'b) -> 'a option -> 'b option
+val option_smart_map : ('a -> 'a) -> 'a option -> 'a option
 val pp_option :
   (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a option -> unit
 val option_mapacc :
@@ -275,4 +291,19 @@ end
 (* file access *)
 val std_resolver :
   ?cwd:string -> paths:string list -> unit ->
-     (?cwd:string -> unit:string -> unit -> string)
\ No newline at end of file
+     (?cwd:string -> unit:string -> unit -> string)
+
+
+type constant = int
+val pp_constant : Format.formatter -> constant -> unit
+val show_constant : constant -> string
+val compare_constant : constant -> constant -> int
+val pp_const : constant spaghetti_printer
+module Constants : sig
+  type t = constant
+  module Map : Map.S with type key = constant
+  module Set : Set.S with type elt = constant
+  val pp : Format.formatter -> t -> unit
+  val show : t -> string
+  val compare : t -> t -> int
+end
\ No newline at end of file
